#!/usr/bin/env python

"""
    bootstrap.py
    
    Create a Python module from the preformatted text contained in the
    xhtmlhook.html document.
"""

from xml.dom import minidom
import os, string, sys

def _indent_docstring(text, cmtuple):

    """
    list = _indent_docstring(text, docstring)
    
    Returned an indented docstring based on the indentation used in the
    text which it precedes.
    """

    line = 0
    indentation = 0
    
    while line < len(text):
        
        this_line = text[line]
        
        spaces = this_line.count(u' ')
        
        if spaces < len(this_line):
        
            # Not all the characters on the line are spaces.
        
            while indentation <= spaces:
            
                if this_line[indentation] != u' ':
                
                    break
                
                indentation = indentation + 1
            
            break
        
        # Try the next line.
        
        line = line + 1
    
    # Wrap the text found to fit within lines of the required width.
    
    leftover = u''
    
    # Add any previous text left over from preceding lines and try to
    # split the line at a space.
    
    cmtype, cmtext = cmtuple
    
    if cmtype == "docstring":
    
        cmwidth = 78 - indentation
    
    else:
    
        cmwidth = 76 - indentation
    
    new_text = []
    leftover = u''
    n = 0
    
    while n < len(cmtext) or leftover != u'':
    
        if n < len(cmtext) and leftover == u'':
        
            # Only read another line if the previous line(s) have been
            # written.
        
            line = cmtext[n]
            n = n + 1
        
        else:
        
            line = leftover
    
        # Wrap the text to the desired line width.
        
        at = len(line)
        
        while at >= cmwidth:
        
            at = line.rfind(u' ', 0, at)
        
        # If no spaces are found then keep the whole line together.
        
        if at == -1: at = len(line)
        
        line, leftover = line[:at], line[at + 1:]
        
        # The comment line contained more than just whitespace so include
        # a leading comment character and a space ("# ").
        
        if cmtype == "comment" and line.count(u' ') < len(line):
        
            line = u'# ' + line
        
        # Add the new line to the list of lines to write to the output
        # list.
        
        new_text.append((indentation * u' ') + line)
    
    # The above loop should catch any left over text so we should be able
    # to just return the list of new lines to the calling method.
    
    return new_text

# Introduce a method for removing single lines from paris of duplicate
# adjacent lines of whitespace.

def _reduce_empty(text):

    """
    list = _reduce_empty(text)
    
    Try to remove duplicate consecutive lines which only consist of
    spaces.
    """
    # Take the first item and place it in a new list. Record the last
    # item added.

    new = text[:1]
    last = text[:1]
    
    # Examine the items in the rest of the list.
    
    for item in text[1:]:
    
        if last is None:
        
            # If there are no restrictions on the next item in the list
            # then add a new item to the list and update the last item
            # variable.
        
            new.append(item)
            last = item
        
        elif item == last:
        
            # If the item is the same as the last item added to the new
            # list then check whether it is just made up of spaces.
        
            if item.count(u' ') < len(item):
            
                # A non-space character was found so add this item to the
                # new list.
            
                new.append(item)
        
        # If all the characters were spaces then nothing is added to the
        # new list and the last item remains the same.
        
        else:
        
            # The last item added to the new list and the current item in
            # this list are not the same. Add the item to the new list
            # and update the last item variable.
        
            new.append(item)
            last = item
    
    # Return the list of new items.
    
    return new


def _return_text(document, in_text = None, listings = None, comments = None):

    """
    listings, comments = return_text(document)
    
    Return a list of tuples, each containing the name of a code block and
    a list of lines in that block.
    The comments list returned contains elements which are lists of lines
    in each comment or docstring found.
    """
    # Create a new list of output and comments for this level of
    # recursion.
    
    if listings == None:
    
        listings = [[None, []]]
    
    if comments is None:
    
        comments = []
    
    for node in document.childNodes:
    
        # Each node is checked to determine whether it has any child
        # nodes.
    
        if node.hasChildNodes():
        
            # If so, then its name and class (if present) are checked
            # against the values we wish to find.
        
            if node.getAttribute(u'class') == u'Python' and \
                in_text is None:
            
                # If the node is a preformatted text node and the class
                # used is for Python then the contents of the relevant
                # child nodes are appended to the list of output. We
                # descend, setting the second parameter to one to
                # indicate that text found beneath this node is to be
                # collected.
            
                listing, cm = _return_text(node, u'Python')
                
                # Join all the text to create a single string and split
                # it at each newline.
                
                text = listing[0][1]
                
                joined = u''.join(text)
                text = joined.split(u'\n')
                
                # If a comment or docstring was being remembered then
                # indent it correctly using the indentation of the first
                # non-empty line as a guide and reset the docstring to
                # None.
                
                for comment in comments:
                
                    listings[-1][1] = listings[-1][1] + _indent_docstring(text, comment)
                    
                comments = []
                
                # Add the list of preformatted text lines to the output
                # list.
                
                listings[-1][1] = listings[-1][1] + text
            
            elif node.getAttribute(u'class') == u'Docstring' and \
                in_text is None:
            
                # For paragraph nodes, check whether the relevant class
                # is used and generate docstrings if it is. Ensure that
                # instances of quotes are escaped and that the level of
                # indentation remains the same.
            
                listing, cm = _return_text(node, u'Docstring')
                
                # The docstring value, ds, is meaningless in this case as
                # all the text for the docstring is contained within an
                # element in the listing list.
                
                text = listing[0][1]
               
                # Join the items in the text list then escape all the
                # quotes used in the complete string.
                
                text = u''.join(text)
                
                # Join together line breaks (from <br/> elements) and
                # carriage returns (generated from newlines in <p>
                # elements) to form a single newline in each case.
                
                pieces = text.split(u'\n\r')
                text = u'\n'.join(pieces)
                
                new_text = []
                
                for c in text:
                
                    if c == u'"':
                        new_text.append(u'\\"')
                    elif c == u'\r':
                        new_text.append(u' ')
                    else:
                        new_text.append(c)
                
                new_text = u''.join(new_text)
                
                # Compile a docstring list from the text returned.
                
                docstring = [u'"""']
                docstring = docstring + new_text.split(u'\n')
                docstring.append(u'"""')
                
                comments.append(("docstring", docstring))
            
            elif node.getAttribute(u'class') == u'Comment' and \
                in_text is None:
            
                # It is useful to be able to include comments in the
                # generated source code, although this is mainly useful
                # for generating a new version of the bootstrap.py file
                # from an XHTML file. Note that we write any docstring
                # being collected since it should appear before comments
                # in the source code.
            
                listing, cm = _return_text(node, u'Comment')
                
                text = listing[0][1]
                
                joined = u''.join(text)
                
                # Join together line breaks (from <br/> elements) and
                # carriage returns (generated from newlines in <p>
                # elements) to form a single newline in each case.
                
                pieces = joined.split(u'\n\r')
                joined = u'\n'.join(pieces)
                
                pieces = joined.split(u'\r')
                joined = u' '.join(pieces)
                
                text = joined.split(u'\n')
                
                # Add the comment lines to the comments list, specifying
                # the appropriate type..
                
                comments.append(("comment", text))
            
            elif node.getAttribute(u'class') == u'Submodule' and in_text is None:
            
                # If we encounter a node with children which has the
                # correct class attribute then collect the textual
                # contents of the child nodes. This does not affect the
                # persistent listings or comments lists until the text
                # has been assembled.
            
                listing, cm = _return_text(node, u'Submodule')
                
                text = listing[0][1]
                
                joined = u''.join(text)
                
                # Start a new listing using the text collected as the
                # name of the listing.
                
                listings.append([joined, []])
            
            else:
            
                # If the node fails to satisfy the criteria then we
                # continue looking for suitable preformatted text in its
                # child nodes, maintaining the comments list.
            
                listings, comments = _return_text(node, in_text, listings, comments)
        
        # For nodes with no child nodes, we are only concerned with those
        # which are located within preformatted areas of the required
        # class.
        
        elif in_text is not None:
        
            # If suitable text has been found the contents of the node
            # are added to the output list.
        
            if node.nodeName == "#text" and in_text == u'Python':
            
                # Suitable text has been found. Add the contents of the
                # node to the output list.
                
                listings[-1][1].append(node.data)
            
            elif node.nodeName == "#text" and (in_text == u'Docstring' or in_text == u'Comment'):
            
                # For paragraph text being collected for comments or
                # docstrings add the contents of the node to the output
                # list.
                
                lines = node.data.split(u'\n')
                
                listings[-1][1].append(u'\r'.join(lines))
            
            elif node.nodeName == "#text" and in_text == u'Submodule':
            
                # For a submodule name declaration, just collect the text
                # as it is for later processing.
                
                listings[-1][1].append(node.data)
            
            # If line breaks are found then insert newlines; this is
            # required because Mozilla appears to find it fashionable to
            # include line break tags in preformatted text rather than
            # actual newlines.
        
            elif node.nodeName == u'br':
            
                # Append a newline to the text collected.
                listings[-1][1].append(u'\n')
    
    # The contents of the output list and the docstring list are
    # returned.
    
    return listings, comments


def main(path, module_name, submodules = None):

    # Look for the xhtmlhook.html document on this path.
    filename = os.path.join(path, module_name + os.extsep + "html")
    
    try:
    
        doc = open(filename, "r")
    
    except IOError:
    
        print "Failed to open the source document: %s" % filename
        sys.exit(1)
    
    # Read and parse the document.
    xhtml = doc.read()
    doc.close()
    
    d = minidom.parseString(xhtml)
    
    # Extract the Python code as a list of unicode strings representing the
    # lines of source code and write it to a file.
    listings = _return_text(d)[0]
    
    if len(listings) > 1:
    
        directory = os.path.join(path, module_name)
        
        try:
        
            os.mkdir(directory)
        
        except OSError:
        
            if not (os.path.exists(directory) and os.path.isdir(directory)):
            
                print "Failed to create package directory: %s" % directory
                sys.exit(1)
        
        for code_name, code_listing in listings:
        
            if code_name is None:
            
                code_name = "__init__"
            
            # If no submodules are specified then extract them all, otherwise
            # only extract the ones listed.
            if submodules is not None and code_name not in submodules:
            
                continue
            
            code_listing = _reduce_empty(code_listing)
            
            text = u'\n'.join(code_listing)
            
            if text != u'' and text[-1] != u'\n':
            
                text = text + u'\n'
            
            #ascii_string = text.encode("ASCII")
            
            filename = os.path.join(directory, code_name + os.extsep + "py")
            
            try:
            
                module = open(filename, "a")
            
            except IOError:
            
                print "Failed to open a file for writing: %s" % filename
                sys.exit(1)
            
            # Write the code to the new file.
            module.write(text)
            module.close()
            
            print "Module written: %s" % filename
    
    else:
    
        filename = os.path.join(path, module_name + os.extsep + "py")
        
        try:
        
            module = open(filename, "w")
        
        except IOError:
        
            print "Failed to open a file for writing: %s" % filename
            sys.exit(1)
        
        # Write the code to the new file.
        module.write(u'\n'.join(listings[0][1]))
        module.close()
        
        print "Success"
        print "Module written: %s" % filename

if __name__ == "__main__":

    syntax = "[module [submodules]]"
    
    # Read the path to this file.
    path = sys.argv[0]
    
    # Extract the path of the directory containing this file.
    path, file = os.path.split(path)
    
    if len(sys.argv) > 3:
    
        print "Too many arguments."
        print
        print "Usage: %s %s" % (file, syntax)
        print
        sys.exit(1)
    
    elif len(sys.argv) == 3:
    
        module_name = sys.argv[1]
        submodules = string.split(sys.argv[2], ",")
    
    elif len(sys.argv) == 2:
    
        module_name = sys.argv[1]
        submodules = None
    
    else:
    
        module_name = "xhtmlhook"
        submodules = None
    
    main(path, module_name, submodules)
    
    # Exit
    sys.exit()
